require 'trellohub/form/card'
require 'trellohub/form/issue'

module Trellohub
  class Form
    include Form::Card
    include Form::Issue

    class << self
      def common_attributes
        %i(
          key
          state
          imported_from
        )
      end

      def origin_attributes
        %i(
          origin_issue
          origin_card
        )
      end

      def array_ext
        <<-METHODS
          def find_by_key(key)
            self.find { |form| form.key == key }
          end
        METHODS
      end

      def with_issues
        @issues ||= self.with_issues!
      end

      def with_issues!
        array = Trellohub.repositories.each.with_object([]) do |repo, forms|
          forms.concat with_issues_on(repo)
        end
        array.instance_eval array_ext
        array
      end

      def with_issues_on(repo)
        repo.issues.each.with_object([]) do |issue, forms|
          form = Trellohub::Form.new
          form.import_issue repo.full_name, issue
          forms << form
        end
      end

      def with_cards
        @cards ||= self.with_cards!
      end

      def with_cards!
        array = Trellohub::Card.all.
          each.with_object([]) do |card, forms|
          form = Trellohub::Form.new
          form.import_card card
          forms << form
        end
        array.instance_eval array_ext
        array
      end

      def compare(base, target)
        return unless base.updated_at < target.updated_at
        type = target.imported_from

        printings = [[
          "#{type} attribute",
          "#{'base'.green} (#{base.imported_from}: #{base.own_key}, #{base.updated_at})",
          "#{'comparison'.yellow} (#{target.imported_from}: #{target.own_key}, #{target.updated_at})"
        ]] if Trellohub.debug

        diff = target.send(:"to_valid_#{type}").each.with_object({}) do |(key, value), hash|
          base_value = base.send(:"to_valid_#{type}")[key]
          hash[key] = value unless value == base_value

          printings << [
            key,
            base_value.to_s.color(value == base_value ? :green : :red),
            value.to_s.yellow
          ] if Trellohub.debug
        end

        if Trellohub.debug && !diff.empty?
          max = printings.max_lengths
          printings.each.with_index(1) do |line, index|
            puts '[DIFF: Update the base]' if index == 1
            puts 3.times.map.with_index { |i| ''.ljust(max[i], '-') }.join(' | ') if index == 2
            puts line.map.with_index { |word, i| "#{word}".ljust(max[i]) }.join(' | ')
          end
        end

        diff unless diff.empty?
      end
    end

    attr_accessor(*self.common_attributes + self.origin_attributes)

    def created_at
      @created_at ||= send(:"#{@imported_from}_#{__method__}")
    end

    def updated_at
      return @issue_updated_at if @imported_from == :issue
      @updated_at ||= card_updated_at || card_created_at
    end

    def closed_at
      return @issue_closed_at if @imported_from == :issue
      @closed_at ||= card_closed_at || card_created_at
    end

    def open?
      @state == 'open'
    end

    def closed?
      @state == 'closed'
    end

    def own_key
      if @imported_from == :issue
        @key
      else
        @card_shortLink
      end
    end

    def to_hash
      Hash[instance_variables.map { |variable|
        next if variable.is_a?(Sawyer::Resource)
        variable_no_at = variable.to_s.gsub('@', '')
        [variable_no_at.to_sym, instance_variable_get(:"#{variable}")]
      }.compact]
    end
  end
end
